/**
 * This work was created by participants in the DataONE project, and is
 * jointly copyrighted by participating institutions in DataONE. For 
 * more information on DataONE, see our web site at http://dataone.org.
 *
 *   Copyright ${year}
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 * 
 * $Id$
 */

package gov.ornl.mercury3.web;

import gov.ornl.mercury3.commands.Configuration;
import gov.ornl.mercury3.commands.OpenSearchBean;
import gov.ornl.mercury3.commands.Search_Terms;
import gov.ornl.mercury3.commands.SolrTransactionDetail;
import gov.ornl.mercury3.commands.Terms;
import gov.ornl.mercury3.services.SolrTransactionService;
import gov.ornl.mercury3.services.SolrTransactionServiceDefault;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FilenameFilter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.validation.ConstraintViolation;
import javax.validation.Validation;
import javax.validation.Validator;
import javax.validation.ValidatorFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.dataone.service.exceptions.BaseException;
import org.dataone.service.exceptions.InvalidRequest;
import org.dataone.service.exceptions.ServiceFailure;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;

public class PilotCatalogController extends SolrClientController {

    Properties mercuryCommonProp = new Properties();

    String webAppRoot = "";

    HashMap<String, Object> hmProps = new HashMap<String, Object>();
    List<String> li = new ArrayList<String>();

    ApplicationContext factory2 = null;
    String metadata_path = "";
    public boolean debug = false;
    String fullViewPage = "";
    private static Log logger = LogFactory.getLog(PilotCatalogController.class);

    public PilotCatalogController() {

        try {

            ApplicationContext factory = new ClassPathXmlApplicationContext(
                    "Mercury3Properties.xml");//

            Configuration cv = (Configuration) factory.getBean("propertiesBean");

            hmProps = cv.getProperties();

            factory2 = new ClassPathXmlApplicationContext(
                    "gov/ornl/mercury3/services/applicationContext.xml");

            metadata_path = (String) hmProps.get("metadata_path");

            fullViewPage = (String) hmProps.get("fullViewPage");

            debug = Boolean.parseBoolean((String) (hmProps.get("debug")));

            System.out.println("In PilotCatalog controller v2  debug = " + debug);

        } catch (Exception ex) {
            ex.printStackTrace();
        }

    }

    /**
     * mimics opensearch but writes to response outputstream and implements
     * dataONE Service Exceptions
     * 
     * @param request
     *            the HttpServletRequest
     * @param response
     *            the HttpServletResponse
     * @param OpenSearchBean
     *            command object containing query parameters
     * 
     * @return void
     * @throws Exception
     * 
     *             &query=carbon&start=0&pageSize=10&stylesheet=
     *             object_collection_xml&sortattribute=default
     * 
     */

    @SuppressWarnings("unchecked")
    public void searchobjects(HttpServletRequest request, HttpServletResponse response,
            OpenSearchBean opensearch) throws Exception {

        StylesheetType acceptHeaderValue = StylesheetType.XML;
        String stylesheet = opensearch.getStylesheet();

        // retrieve the accept headers so as to determine output format
        // should be the same as stylesheet, but in case there is an
        // exception raised before stylesheet parsing, it would
        // be good to know how to return serialized exception
        // the default is XML
        Enumeration<String> headerNames = request.getHeaderNames();
        while (headerNames.hasMoreElements()) {
            String headerName = headerNames.nextElement();
            if (headerName.equalsIgnoreCase("Accept")) {
                if (request.getHeader(headerName).contains("json")) {
                    acceptHeaderValue = StylesheetType.JSON;
                } else if (request.getHeader(headerName).contains("csv")) {
                    acceptHeaderValue = StylesheetType.CSV;
                }
            }
        }
        if (debug) {
            logger.debug("Headers!\n");
            while (headerNames.hasMoreElements()) {
                String headerName = headerNames.nextElement();
                logger.debug(headerName + ": " + request.getHeader(headerName));
            }
            // logger.debug(request.getCharacterEncoding() + " : " +
            // request.getContentType() + " : " + request.getContentLength());
            logger.debug("Attributes!\n");
            Enumeration<String> attributeNames = request.getAttributeNames();
            while (attributeNames.hasMoreElements()) {
                String attributeName = attributeNames.nextElement();
                logger.debug(attributeName + ": " + request.getAttribute(attributeName));
            }
            logger.debug("Parameters!\n");
            Enumeration<String> parameterNames = request.getParameterNames();
            while (parameterNames.hasMoreElements()) {
                String parameterName = parameterNames.nextElement();
                logger.debug(parameterName + ": " + request.getParameter(parameterName));
            }
        }
        StylesheetType stylesheetType = StylesheetType.HTML;
        try {

            if ((null == stylesheet) || (stylesheet.length() == 0)) {
                response.setStatus(response.SC_BAD_REQUEST);
                BaseException invalidRequest = (BaseException) new InvalidRequest("4300",
                        "Xsl stylesheet is missing for this request: " + request.getServletPath()
                                + ": " + request.getQueryString());
                throw invalidRequest;
            }

            // Confirm that the stylesheet name matches a file in
            // gov.ornl.mercury3.services
            StylesheetFilter stylesheetFilter = new StylesheetFilter(stylesheet, "xsl");
            File path = factory2.getResource("classpath:gov/ornl/mercury3/services").getFile();

            File[] files = path.listFiles(stylesheetFilter);
            logger.debug("Looking for a match in " + path.getAbsolutePath() + "\n");
            if (files.length == 0) {
                response.setStatus(response.SC_BAD_REQUEST);
                BaseException invalidRequest = (BaseException) new InvalidRequest("4300",
                        "Xsl stylesheet name: " + stylesheet + " is misspelled or is unavailable: "
                                + request.getServletPath() + request.getQueryString());
                throw invalidRequest;
            }
            if (stylesheet.endsWith("json")) {
                stylesheetType = StylesheetType.JSON;
            } else if (stylesheet.endsWith("xml")) {
                stylesheetType = StylesheetType.XML;
            } else if (stylesheet.endsWith("csv")) {
                stylesheetType = StylesheetType.CSV;
            } else {
                response.setStatus(response.SC_BAD_REQUEST);
                BaseException invalidRequest = (BaseException) new InvalidRequest("4300",
                        "Xsl stylesheet name: " + stylesheet
                                + " does not contain a supported mimetype: "
                                + request.getServletPath() + request.getQueryString());
                throw invalidRequest;
            }
            logger.debug("\nquery: " + opensearch.getQuery() + "\n" + "start: "
                    + opensearch.getStart() + "\n" + "pageSize: " + opensearch.getPageSize() + "\n"
                    + "stylesheet: " + opensearch.getStylesheet() + "\n" + "sortattribute: "
                    + opensearch.getSortattribute() + "\n" + "startTime: "
                    + opensearch.getStartTime() + "\n" + "endTime: " + opensearch.getEndTime());

            ValidatorFactory factory = Validation.buildDefaultValidatorFactory();
            Validator validator = factory.getValidator();
            Set<ConstraintViolation<OpenSearchBean>> violations = validator.validate(opensearch);
            for (ConstraintViolation<OpenSearchBean> violation : violations) {
                InvalidRequest invalidRequest = new InvalidRequest("4300", "Invalid value for: '"
                        + violation.getPropertyPath().toString() + "': " + violation.getMessage());
                throw invalidRequest;
            }

            SolrTransactionService solrTransactionServices = new SolrTransactionServiceDefault();
            Search_Terms st = new Search_Terms();
            try {
                logger.debug("get_base_st");
                st = get_base_st(request, response);
                logger.debug("null != opensearch.getStartTime()");
                if ((null != opensearch.getStartTime()) && (opensearch.getStartTime().length() > 0)) {
                    Terms temporal1 = new Terms();
                    temporal1.setAttribute("beginDate");
                    temporal1.setOperation("during");
                    temporal1.setValue(opensearch.getStartTime());
                    st.setTerm4(temporal1);
                }
                logger.debug("null != opensearch.getEndTime()");
                if ((null != opensearch.getEndTime()) && (opensearch.getEndTime().length() > 0)) {
                    Terms temporal2 = new Terms();
                    temporal2.setAttribute("endDate");
                    temporal2.setOperation("during");
                    temporal2.setValue(opensearch.getEndTime());
                    st.setTerm5(temporal2);
                }
                if ((null != opensearch.getSortattribute())
                        && (opensearch.getSortattribute().length() > 0)) {
                    st.setSortattribute(opensearch.getSortattribute());
                } else {
                    st.setSortattribute("default");
                }
                logger.debug("st.setFeedValue");
                st.setFeedValue(opensearch.getStylesheet());

                st.setStart(Integer.parseInt(opensearch.getStart()));
                st.setSortattribute(opensearch.getSortattribute());
                st.setPageSize(Integer.parseInt(opensearch.getPageSize()));
                Terms term1 = new Terms();
                term1.setValue(opensearch.getQuery());
                term1.setAttribute("fullText");
                st.setTerm1(term1);

            } catch (Exception exc) {

                ByteArrayOutputStream printExceptionTraceStream = new ByteArrayOutputStream();
                PrintWriter printExceptionTraceWriter = new PrintWriter(printExceptionTraceStream);
                exc.printStackTrace(printExceptionTraceWriter);
                printExceptionTraceWriter.flush();
                printExceptionTraceStream.flush();
                response.setStatus(response.SC_INTERNAL_SERVER_ERROR);
                BaseException serviceException = (BaseException) new ServiceFailure("4310",
                        "Class: " + exc.toString() + " Cause: " + exc.getCause() + " Message: "
                                + exc.getMessage() + "\n" + printExceptionTraceStream.toString());
                throw serviceException;
            }
            logger.debug("st.setRSS");
            st.setRSS(true);

            try {
                logger.debug("getSolrTransactionDetail");
                SolrTransactionDetail solrTransactionDetail = solrTransactionServices
                        .getSolrTransactionDetail(st, null);
                if (solrTransactionDetail.getRes().getRss1().isEmpty()) {
                    logger.error("getRss returned an Empty String. No Result returned from Solr");
                    Exception e = new Exception(
                            "getRss returned an Empty String. No Result returned from Solr");
                    throw e;
                }
                ServletOutputStream outputStream = response.getOutputStream();
                String rssResult = solrTransactionDetail.getRes().getRss1();
                logger.debug("print solrTransactionDetail Rss1 :" + rssResult);
                byte[] resultArray = rssResult.getBytes("UTF-8");

                outputStream.write(resultArray, 0, resultArray.length);

                outputStream.flush();
                outputStream.close();
            } catch (Exception ex) {

                response.setStatus(response.SC_INTERNAL_SERVER_ERROR);
                BaseException serviceException = (BaseException) new ServiceFailure("4310",
                        ex.getCause() + " " + ex.getMessage());
                throw serviceException;
            }

        } catch (BaseException baseException) {
            // baseException.printStackTrace();

            ServletOutputStream outputStream = response.getOutputStream();
            byte[] exceptionArray;
            switch (acceptHeaderValue) {
            case CSV: {
                exceptionArray = baseException.serialize(BaseException.FMT_HTML).getBytes("UTF-8");
                outputStream.write(exceptionArray, 0, exceptionArray.length);
                break;
            }
            case JSON: {
                exceptionArray = baseException.serialize(BaseException.FMT_JSON).getBytes("UTF-8");
                outputStream.write(exceptionArray, 0, exceptionArray.length);
                break;
            }
            case XML: {
                exceptionArray = baseException.serialize(BaseException.FMT_XML).getBytes("UTF-8");
                outputStream.write(exceptionArray, 0, exceptionArray.length);
                break;
            }
            case HTML: {
                exceptionArray = baseException.serialize(BaseException.FMT_HTML).getBytes("UTF-8");
                outputStream.write(exceptionArray, 0, exceptionArray.length);
                break;
            }
            }
            outputStream.flush();
            outputStream.close();
        } catch (Exception e) {
            // e.printStackTrace();

            response.setStatus(response.SC_INTERNAL_SERVER_ERROR);
            ServiceFailure serviceException = new ServiceFailure("4310", e.getCause() + " "
                    + e.getMessage());
            ServletOutputStream outputStream = response.getOutputStream();
            byte[] exceptionArray = null;
            switch (acceptHeaderValue) {
            case CSV:
                exceptionArray = serviceException.serialize(BaseException.FMT_HTML).getBytes(
                        "UTF-8");
            case JSON:
                exceptionArray = serviceException.serialize(BaseException.FMT_JSON).getBytes(
                        "UTF-8");
            case XML:
                exceptionArray = serviceException.serialize(BaseException.FMT_XML)
                        .getBytes("UTF-8");
            case HTML:
                exceptionArray = serviceException.serialize(BaseException.FMT_HTML).getBytes(
                        "UTF-8");
            }
            outputStream.write(exceptionArray, 0, exceptionArray.length);
            outputStream.flush();
            outputStream.close();
        }
        return;
    }

    private enum StylesheetType {
        XML, JSON, CSV, HTML
    }

    private class StylesheetFilter implements FilenameFilter {
        private String nameStart;

        private String extension;

        public StylesheetFilter(String nameStart, String extension) {
            this.nameStart = nameStart;
            this.extension = extension;
        }

        public boolean accept(File directory, String filename) {
            boolean fileOK = true;

            if (nameStart != null) {
                fileOK &= filename.startsWith(nameStart);
            }

            if (extension != null) {
                fileOK &= filename.endsWith('.' + extension);
            }
            return fileOK;
        }
    }
}
